/*
 * Procedures for converting a logic specification from an expression
 * tree into array form.
 */
#include "logsyn.h"
#include "logsyn.tab.h"
#include <stdio.h>

short *bitnum_table;	/* stores mapping from signals to bit numbers */
int num_inputs, num_vars;

/*
 * In the following procedures, we use the following representation
 * in the output columns of the arrays: X for elements of the ON array,
 * 1 for elements of the DC (don't care) array.  This is so that
 * 1 & % = %, % | 1 = 1, etc.  Elements of the Z array are not explicitly
 * represented.
 */

/*
 * Convert an expression into array form.
 * The order and number of inputs are given in the group `ins'.
 */
struct array *
expr_array(exp, ins, pol)
struct expr *exp, *ins;
int pol;
{
	struct expr *ep, *p;
	struct signal *sp;
	struct array *ap, *oap;
	int i, bnum;

	exp = make_group(exp);
	ins = make_group(ins);
	printf("inputs: ");
	pexpr(ins);
	printf("\n");
	num_inputs = ins->e_nel;		/* # inputs */
	num_vars = num_inputs + exp->e_nel;	/* # inputs + outputs */
	bitnum_table = (short *)malloc(num_signals * sizeof(short));
	for( i = 0; i < num_signals; ++i )
		bitnum_table[i] = -1;	/* => signal not involved */
	bnum = 0;
	for( p = ins->e_first; p != NULL; p = p->e_right ){
		ep = p->e_left;
		if( ep->e_op != SIGNAL )
			cverror("input %d is not a signal", bnum+1);
		else {
			sp = ep->e_signal;
			i = sp->s_num + ep->e_index - sp->s_ix.i_minix;
			bitnum_table[i] = bnum;
		}
		++bnum;
	}

	ap = newarray(num_inputs, num_vars);
	for( p = exp->e_first; p != NULL; p = p->e_right ){
		ep = p->e_left;
		oap = conv_func(ep, bnum, pol);
		chcol(oap, bnum, Z, X, O);	/* undo outrep */
		append(ap, oap);
		++bnum;
	}
	free(bitnum_table);
	return ap;
}

/* Set up bitnum_table for a series of conversions */
conv_inputs(eap, noin)
struct expr **eap;
int noin;
{
	int i, j;
	struct expr *ep;
	struct signal *sp;

	num_inputs = noin;
	bitnum_table = (short *)malloc(num_signals * sizeof(short));
	for( i = 0; i < num_signals; ++i )
		bitnum_table[i] = -1;	/* => signal not involved */
	for( i = 0; i < noin; ++i ){
		ep = eap[i];
		if( ep != NULL ){
			if( ep->e_op != SIGNAL )
				cverror("input %d is not a signal", i+1);
			else {
				sp = ep->e_signal;
				j = sp->s_num + ep->e_index - sp->s_ix.i_minix;
				bitnum_table[j] = i;
			}
		}
	}
}

/* Convert an expression (not a group) to an array with 1 output
   and minimize it. */
struct array *
conv_1_expr(ep, pol)
struct expr *ep;
int pol;
{
	struct array *oap;
	extern int pdebug;

	num_vars = num_inputs + 1;
	oap = conv_func(ep, num_inputs, pol);
	chcol(oap, num_inputs, Z, X, O);	/* undo outrep */
#if 0
	if( pdebug ){
		printf("function:\n");
		prary(oap, stdout);
	}
#endif
	return oap;
}

/* Finished doing conversions */
conv_finish()
{
	free(bitnum_table);
}

int
n_terms(ap)
struct array *ap;
{
	int n;
	register struct cube *cp;

	n = 0;
	for( cp = ap->a_first; cp != NULL; cp = cp->c_next )
		++n;
	return n;
}

struct cube *
make_glfree(grp)
    struct expr *grp;
{
    struct cube *cp;

    cp = newcube(num_inputs);
    if (grp == NULL) {
	setcube(cp, 0, num_inputs, 1);
    } else {
	setcube(cp, 0, num_inputs, 0);
	make_glfree_sub(cp, grp);
    }
    return cp;
}

make_glfree_sub(cp, ep)
    struct cube *cp;
    struct expr *ep;
{
    int op, i;
    struct signal *sp;

    op = ep->e_op;
    switch (op) {
    case SIGNAL:
	sp = ep->e_signal;
	i = bitnum_table[sp->s_num + ep->e_index - sp->s_ix.i_minix];
	if (i >= 0)
	    cp->c_logic[i] = 1;
	break;
    case ',':
	make_glfree_sub(cp, ep->e_left);
	make_glfree_sub(cp, ep->e_right);
	break;
    default:
	yyerror("can't cope with op %d in hazard_free list", op);
    }
}

short invpol[4] = { N, O, Z, X };	/* inverted bits */
short outrep[4] = { N, N, X, O };	/* representation in output cols */

struct cube *
make_cube(onum, oval)
int onum, oval;
{
	struct cube *cp;

	cp = newcube(num_vars);
	setcube(cp, 0, num_inputs, X);
	setcube(cp, num_inputs, num_vars-num_inputs, Z);
	cp->c_logic[onum] = outrep[oval];
	return cp;
}

/* Convert the function representing output `onum'. */
/* `pol' = O to get function itself, Z to get inverse. */
struct array *
conv_func(ep, onum, pol)
struct expr *ep;
int onum, pol;
{
	struct array *ap, *aq, *ar, *as;
	struct cube *cp, *x;
	struct signal *sp;
	int i, b, inum, op, ipol, np;
	struct expr *eq;
	struct array **papo, **papz;
	extern char *signal_name();

	if( ep == NULL ){
		ap = newarray(num_inputs, num_vars);
		if( pol == Z )
			add(ap, make_cube(onum, O));
		return ap;
	}
	op = ep->e_op;
	ipol = X - pol;
	switch( op ){
	case SIGNAL:
		ap = newarray(num_inputs, num_vars);
		sp = ep->e_signal;
		i = sp->s_num + ep->e_index - sp->s_ix.i_minix;
		inum = bitnum_table[i];
		if( inum == -1 ){
			yyerror("signal %s not in input list",
				signal_name(ep));
			bitnum_table[i] = -2;	/* don't shriek next time */
		}
		if( inum < 0 )
			break;
		cp = make_cube(onum, O);
		cp->c_logic[inum] = pol;
		add(ap, cp);
		break;
	case BIT:
		ap = newarray(num_inputs, num_vars);
		b = ep->e_bval;
		if( pol == Z )
			b = invpol[b];
		if( b >= O ){
			cp = make_cube(onum, b);
			add(ap, cp);
		}
		break;
	case '&':
	case '|':
	case LESS:
	case GTR:
	case LESSEQ:
	case GTREQ:
	case '#':
		ap = conv_func(ep->e_left, onum,
			(op == LESS || op == LESSEQ? ipol: pol));
		aq = conv_func(ep->e_right, onum,
			(op == GTR || op == GTREQ || op == '#'? ipol: pol));
		if( (op == '|' || op == LESSEQ || op == GTREQ)
		     != (pol == O) ){
			/* AND function */
			hass(ap, intersect(ap, aq));
		} else {
			/* OR function */
			append(ap, aq);
		}
		break;
	case '~':
		ap = conv_func(ep->e_left, onum, ipol);
		break;
	case '^':
	case NOTEQUAL:
	case EQUAL:
		ap = conv_func(ep->e_left, onum, Z);
		aq = conv_func(ep->e_right, onum, (op != EQUAL? pol: ipol));
		hass(ap, intersect(ap, aq));
		aq = conv_func(ep->e_left, onum, O);
		ar = conv_func(ep->e_right, onum, (op != EQUAL? ipol: pol));
		hass(aq, intersect(aq, ar));
		append(ap, aq);
		break;
	case IF:
		eq = ep->e_right;	/* ELSE node */
		aq = conv_func(ep->e_left, onum, O);
		ar = conv_func(eq->e_left, onum, pol);
		ap = intersect(aq, ar);
		release(aq);
		aq = conv_func(ep->e_left, onum, Z);
		as = conv_func(eq->e_right, onum, pol);
		hass(aq, intersect(aq, as));
		hass(ar, intersect(ar, as));
		append(ap, aq);
		append(ap, ar);
		release(as);
		break;
	case ARRAY:
		np = ep->e_left->e_nel;
                papo = (struct array **)alloc(np * sizeof(struct array *));
		papz = (struct array **)alloc(np * sizeof(struct array *));
		eq = ep->e_left->e_first;
		for( i = 0; i < np; ++i ){
			papo[i] = conv_func(eq->e_left, onum, O);
			papz[i] = conv_func(eq->e_left, onum, Z);
			eq = eq->e_right;
		}
		ap = newarray(num_inputs, num_vars);
		as = (pol == Z)? inv_array(ep->e_array): ep->e_array;
		for( x = as->a_first; x != NULL; x = x->c_next ){
			aq = newarray(num_inputs, num_vars);
			cp = make_cube(onum, x->c_logic[np]);
			add(aq, cp);
			for( i = 0; i < np; ++i ){
				switch( x->c_logic[i] ){
				case N:
					release(aq);	/* make null */
					aq = newarray(num_inputs, num_vars);
					break;
				case Z:
					hass(aq, intersect(aq, papz[i]));
					break;
				case O:
					hass(aq, intersect(aq, papo[i]));
					break;
				}
			}
			append(ap, aq);
		}
		for( i = 0; i < np; ++i ){
			release(papo[i]);
			release(papz[i]);
		}
		if( pol == Z )
			release(as);
		free(papo);
		free(papz);
		break;
	default:
		fprintf(stderr, "Can't convert operator ");
		fprop(stderr, ep->e_op);
		fprintf(stderr, "\n");
		ap = newarray(num_inputs, num_vars);
		break;
	}
	absorb(ap);
	return ap;
}

cverror(s, p)
char *s;
{
	fprintf(stderr, s, p);
	fprintf(stderr, "\n");
}

